@@ -11,3 +11,6 @@ AUTO_PING = get_settings_value('AUTO_PING', True)
 AUTO_CONFIG_CHECK = get_settings_value('AUTO_DEVICE_CONFIG_CHECK', True)
 MANAGEMENT_IP_ONLY = get_settings_value('MANAGEMENT_IP_ONLY', True)
 PING_CHECK_CONFIG = get_settings_value('PING_CHECK_CONFIG', {})
+CONFIG_CHECK_INTERVAL = int(
+    get_settings_value('CONFIG_CHECK_INTERVAL', 5)
+)  # in minutes
