@@ -1,5 +1,7 @@
 import json
 import logging
+from collections import OrderedDict
+from copy import deepcopy
 from datetime import date, datetime, timedelta
 
 from dateutil.parser import parse as parse_date
@@ -13,6 +15,7 @@ from django.db import models
 from django.utils import timezone
 from django.utils.text import slugify
 from django.utils.translation import gettext_lazy as _
+from jsonfield import JSONField
 from openwisp_notifications.signals import notify
 from pytz import timezone as tz
 from pytz import utc
@@ -50,12 +53,35 @@ class AbstractMetric(TimeStampedEditableModel):
     )
     object_id = models.CharField(max_length=36, db_index=True, blank=True)
     content_object = GenericForeignKey('content_type', 'object_id')
+    main_tags = JSONField(
+        _('main tags'),
+        default=dict,
+        blank=True,
+        load_kwargs={'object_pairs_hook': OrderedDict},
+        dump_kwargs={'indent': 4},
+        db_index=True,
+    )
+    extra_tags = JSONField(
+        _('extra tags'),
+        default=dict,
+        blank=True,
+        load_kwargs={'object_pairs_hook': OrderedDict},
+        dump_kwargs={'indent': 4},
+    )
     # NULL means the health has yet to be assessed
     is_healthy = models.BooleanField(default=None, null=True, blank=True, db_index=True)
+    # Like "is_healthy", but respects tolerance of alert settings
+    is_healthy_tolerant = models.BooleanField(default=None, null=True, blank=True)
 
     class Meta:
         abstract = True
-        unique_together = ('key', 'field_name', 'content_type', 'object_id')
+        unique_together = (
+            'key',
+            'field_name',
+            'content_type',
+            'object_id',
+            'main_tags',
+        )
 
     def __str__(self):
         obj = self.content_object
@@ -64,6 +90,11 @@ class AbstractMetric(TimeStampedEditableModel):
         model_name = obj.__class__.__name__
         return '{0} ({1}: {2})'.format(self.name, model_name, obj)
 
+    def __setattr__(self, attrname, value):
+        if attrname in ['main_tags', 'extra_tags']:
+            value = self._sort_dict(value)
+        return super().__setattr__(attrname, value)
+
     def clean(self):
         if (
             self.field_name == 'value'
@@ -93,11 +124,17 @@ class AbstractMetric(TimeStampedEditableModel):
         if 'key' in kwargs:
             kwargs['key'] = cls._makekey(kwargs['key'])
         try:
-            lookup_kwargs = kwargs.copy()
+            lookup_kwargs = deepcopy(kwargs)
             if lookup_kwargs.get('name'):
                 del lookup_kwargs['name']
+            extra_tags = lookup_kwargs.pop('extra_tags', {})
             metric = cls.objects.get(**lookup_kwargs)
             created = False
+
+            if extra_tags != metric.extra_tags:
+                metric.extra_tags.update(kwargs['extra_tags'])
+                metric.extra_tags = cls._sort_dict(metric.extra_tags)
+                metric.save()
         except cls.DoesNotExist:
             metric = cls(**kwargs)
             metric.full_clean()
@@ -107,7 +144,7 @@ class AbstractMetric(TimeStampedEditableModel):
 
     @property
     def codename(self):
-        """ identifier stored in timeseries db """
+        """identifier stored in timeseries db"""
         return self._makekey(self.name)
 
     @property
@@ -126,18 +163,34 @@ class AbstractMetric(TimeStampedEditableModel):
     # TODO: This method needs to be refactored when adding the other db
     @staticmethod
     def _makekey(value):
-        """ makes value suited for InfluxDB key """
+        """makes value suited for InfluxDB key"""
         value = value.replace('.', '_')
         return slugify(value).replace('-', '_')
 
     @property
     def tags(self):
+        tags = {}
         if self.content_type and self.object_id:
-            return {
-                'content_type': self.content_type_key,
-                'object_id': str(self.object_id),
-            }
-        return {}
+            tags.update(
+                {
+                    'content_type': self.content_type_key,
+                    'object_id': str(self.object_id),
+                }
+            )
+        if self.main_tags:
+            tags.update(self.main_tags)
+        if self.extra_tags:
+            tags.update(self.extra_tags)
+        return tags
+
+    @staticmethod
+    def _sort_dict(dict_):
+        """
+        ensures the order of the keys in the dict not random
+        """
+        if not isinstance(dict_, OrderedDict):
+            return OrderedDict(sorted(dict_.items()))
+        return dict_
 
     @property
     def content_type_key(self):
@@ -154,52 +207,116 @@ class AbstractMetric(TimeStampedEditableModel):
             return parse_date(time)
         return time
 
+    def _set_is_healthy(self, alert_settings, value):
+        """
+        Sets the value of "is_healthy" field if "value"
+        crosses threshold defined in "alert_settings".
+        Returns "True" if "is_healthy" field is changed.
+        Otherwise, returns "None".
+
+        This method does not take into account the alert
+        settings tolerance, which is done by
+        "_set_is_healthy_tolerant" method.
+        """
+        crossed = alert_settings._value_crossed(value)
+        if (not crossed and self.is_healthy) or (crossed and self.is_healthy is False):
+            return
+        # problem: not within threshold limit
+        elif crossed and self.is_healthy in [True, None]:
+            self.is_healthy = False
+        # ok: returned within threshold limit
+        elif not crossed and self.is_healthy in [False, None]:
+            self.is_healthy = True
+        return True
+
+    def _set_is_healthy_tolerant(
+        self, alert_settings, value, time, retention_policy, send_alert
+    ):
+        """
+        Sets the value of "is_tolerance_healthy" if "value"
+        crosses the threshold for more than the amount of seconds
+        defined in the alert_settings "tolerance" field.
+        It also sends the notification if required.
+        Returns "None" if value of "is_healthy_tolerant" is unchanged.
+        Returns "True" if it is the first metric write within threshold.
+        Returns "False" in other cases.
+
+        This method is similar to "_set_is_healthy" but it takes into
+        account the alert settings tolerance so it's slightly different
+        and more complex.
+        """
+        time = self._get_time(time)
+        crossed = alert_settings._is_crossed_by(value, time, retention_policy)
+        first_time = False
+        # situation has not changed
+        if (not crossed and self.is_healthy_tolerant) or (
+            crossed and self.is_healthy_tolerant is False
+        ):
+            return
+        # problem: not within threshold limit
+        elif crossed and self.is_healthy_tolerant in [True, None]:
+            if self.is_healthy_tolerant is None:
+                first_time = True
+            self.is_healthy_tolerant = False
+            notification_type = f'{self.configuration}_problem'
+        # ok: returned within threshold limit
+        elif not crossed and self.is_healthy_tolerant is False:
+            self.is_healthy_tolerant = True
+            notification_type = f'{self.configuration}_recovery'
+        # First metric write within threshold
+        elif not crossed and self.is_healthy_tolerant is None:
+            self.is_healthy_tolerant = True
+            first_time = True
+
+        # If we got to this point, it means we have to send an alert,
+        # because the metric has been crossed for more than the
+        # tolerated amount of time. There's one exception though:
+        # if the device is new, its status will be unknown and the metric
+        # will become healthy for the first time, in this case we do not need
+        # to send an alert.
+        if (
+            not (first_time and self.is_healthy_tolerant)
+            and alert_settings.is_active
+            and send_alert
+        ):
+            self._notify_users(notification_type, alert_settings)
+        return first_time
+
     def check_threshold(self, value, time=None, retention_policy=None, send_alert=True):
         """
-        checks if the threshold is crossed
-        and notifies users accordingly
+        Checks if the threshold is crossed and notifies users accordingly
         """
         try:
             alert_settings = self.alertsettings
         except ObjectDoesNotExist:
             return
-        time = self._get_time(time)
-        crossed = alert_settings._is_crossed_by(value, time, retention_policy)
-        first_time = False
-        # situation has not changed
-        if (not crossed and self.is_healthy) or (crossed and self.is_healthy is False):
+        is_healthy_changed = self._set_is_healthy(alert_settings, value)
+        tolerance_healthy_changed_first_time = self._set_is_healthy_tolerant(
+            alert_settings, value, time, retention_policy, send_alert
+        )
+        is_healthy_tolerant_changed = tolerance_healthy_changed_first_time is not None
+        # Do nothing if none of the fields changed.
+        if not is_healthy_changed and not is_healthy_tolerant_changed:
             return
-        # problem: not within threshold limit
-        elif crossed and self.is_healthy in [True, None]:
-            if self.is_healthy is None:
-                first_time = True
-            self.is_healthy = False
-            notification_type = f'{self.configuration}_problem'
-        # ok: returned within threshold limit
-        elif not crossed and self.is_healthy is False:
-            self.is_healthy = True
-            notification_type = f'{self.configuration}_recovery'
-        # First metric write within threshold
-        elif not crossed and self.is_healthy is None:
-            self.is_healthy = True
-            first_time = True
-        self.save()
+        update_fields = []
+        if is_healthy_changed:
+            update_fields.append('is_healthy')
+        if is_healthy_tolerant_changed:
+            update_fields.append('is_healthy_tolerant')
+        self.save(update_fields=update_fields)
         threshold_crossed.send(
             sender=self.__class__,
             alert_settings=alert_settings,
             metric=self,
             target=self.content_object,
-            first_time=first_time,
+            first_time=tolerance_healthy_changed_first_time,
+            tolerance_crossed=is_healthy_tolerant_changed,
         )
-        # First metric write and within threshold, do not raise alert
-        if first_time and self.is_healthy:
-            return
-        if alert_settings.is_active and send_alert:
-            self._notify_users(notification_type, alert_settings)
 
     def write(
         self,
         value,
+        current=False,
         time=None,
         database=None,
         check=True,
@@ -207,20 +324,27 @@ class AbstractMetric(TimeStampedEditableModel):
         retention_policy=None,
         send_alert=True,
     ):
-        """ write timeseries data """
+        """write timeseries data"""
         values = {self.field_name: value}
         if extra_values and isinstance(extra_values, dict):
             for key in extra_values.keys():
                 if not self.related_fields or key not in self.related_fields:
                     raise ValueError(f'"{key}" not defined in metric configuration')
             values.update(extra_values)
-        signal_kwargs = dict(sender=self.__class__, metric=self, values=values)
+        signal_kwargs = dict(
+            sender=self.__class__,
+            metric=self,
+            values=values,
+            time=time,
+            current=current,
+        )
         pre_metric_write.send(**signal_kwargs)
         options = dict(
             tags=self.tags,
             timestamp=time or timezone.now(),
             database=database,
             retention_policy=retention_policy,
+            current=current,
         )
         # check can be disabled,
         # mostly for automated testing and debugging purposes
@@ -235,13 +359,13 @@ class AbstractMetric(TimeStampedEditableModel):
         timeseries_write.delay(name=self.key, values=values, **options)
 
     def read(self, **kwargs):
-        """ reads timeseries data """
+        """reads timeseries data"""
         return timeseries_db.read(
             key=self.key, fields=self.field_name, tags=self.tags, **kwargs
         )
 
     def _notify_users(self, notification_type, alert_settings):
-        """ creates notifications for users """
+        """creates notifications for users"""
         opts = dict(sender=self, type=notification_type, action_object=alert_settings)
         if self.content_object is not None:
             opts['target'] = self.content_object
@@ -289,13 +413,27 @@ class AbstractChart(TimeStampedEditableModel):
     def type(self):
         return self.config_dict['type']
 
+    @property
+    def fill(self):
+        return self.config_dict.get('fill')
+
+    @property
+    def xaxis(self):
+        return self.config_dict.get('xaxis', {})
+
+    @property
+    def yaxis(self):
+        return self.config_dict.get('yaxis', {})
+
     @property
     def label(self):
         return self.config_dict.get('label') or self.title
 
     @property
     def description(self):
-        return self.config_dict['description'].format(metric=self.metric)
+        return self.config_dict['description'].format(
+            metric=self.metric, **self.metric.tags
+        )
 
     @property
     def title(self):
@@ -378,7 +516,11 @@ class AbstractChart(TimeStampedEditableModel):
         params = dict(field_name=m.field_name, key=m.key, time=self._get_time(time))
         if m.object_id:
             params.update(
-                {'content_type': m.content_type_key, 'object_id': m.object_id}
+                {
+                    'content_type': m.content_type_key,
+                    'object_id': m.object_id,
+                    **m.tags,
+                }
             )
         return params
 
@@ -468,7 +610,7 @@ class AbstractChart(TimeStampedEditableModel):
         """
         rounds value if it makes sense
         """
-        control = 1.0 / 10 ** decimal_places
+        control = 1.0 / 10**decimal_places
         if value < control:
             decimal_places += 2
         return round(value, decimal_places)
@@ -611,7 +753,7 @@ class AbstractAlertSettings(TimeStampedEditableModel):
                         return value_crossed
                     # otherwise, the results are flapping, the situation has not changed
                     # we will return a value that will not trigger changes
-                    return not self.metric.is_healthy
+                    return not self.metric.is_healthy_tolerant
                 # otherwise keep looking back
                 continue
             # the search has not yielded any conclusion
